#include "logger.h"
#include "tokeniser.h"
#include "util.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <token.h>
#include <dynarray.h>
#include <parser/parser.h>
#include <parser/precomp.h>
#include <parser/tokcmp.h>
#include <parser/ast.h>

int _parse(parser_state_t* state);
ast_op_t* parse_fnc_body(parser_state_t* state);

size_t find_mem_id(parser_state_t* state, const char* name) {
    for (int i = 0; i < dynarray_length(state->prog.memories); i++) {
        if (strcmp(state->prog.memories[i].name, name) == 0) {
            return i;
        }
    }
    return -1;
}

size_t find_const_id(parser_state_t* state, const char* name) {
    for (int i = 0; i < dynarray_length(state->prog.const_vars); i++) {
        if (strcmp(state->prog.const_vars[i].name, name) == 0) {
            return i;
        }
    }
    return -1;
}

char* find_func(parser_state_t* state, const char* name) {
    for (int i = 0; i < dynarray_length(state->prog.funcs); i++) {
        if (strcmp(state->prog.funcs[i].name, name) == 0) {
            return (char*)name;
        }
    }
    return NULL;
}

program_t* parse(token_t* tokens) {
    // reverses tokens so its way easier to parse;
    log_debug(NULL, "%zu tokens", dynarray_length(tokens));
    tokens = dynarray_reverse((void*)tokens);
    static parser_state_t state = {0};
    log_debug(NULL, "%zu tokens", dynarray_length(tokens));
    state.tokens = tokens;
    state.prog.const_vars = dynarray_create(const_t);
    state.prog.memories = dynarray_create(memory_t);
    state.prog.funcs = dynarray_create(function_t);

    if (_parse(&state) != 0) {
        return NULL;
    }
    return &state.prog;
}

int _parse(parser_state_t* state) {
    log_debug(&state->curr_tok.loc, "Starting parsing (%zu tokens)", dynarray_length(state->tokens));
    while (dynarray_length(state->tokens)) {
        dynarray_pop(state->tokens, &state->curr_tok);
        //log_debug(&state->curr_tok.loc, "Found Token");
        if (state->curr_tok.type != TT_KW) {
            const char* dbg_s = get_tok_str_dbg(&state->curr_tok);
            log_error(&state->curr_tok.loc, "Invalid word, expected Keyword, got %s.", dbg_s);
            free((void*)dbg_s);
        }
        //log_debug(&state->curr_tok.loc, "Its a KW");
        switch (state->curr_tok.kw_type) {
        case (KW_INCLUDE): {
            expect_token_type(state, TT_PUSH_STR);
            const char* code = read_to_string(state->curr_tok.str_v);
            if (!code) {
                return 1;
            }
            token_t* tokens = tokenise_string((char*)state->curr_tok.str_v, (char*)code);
            if (!tokens) {
                return 1;
            }
            if (_parse(state) != 0) {
                return 1;
            }
        }; break;
        case (KW_CONST): {
            token_t* tokens = dynarray_create(token_t);
            token_t name = *expect_token_type(state, TT_IDENT);
            while (test_token_type_ex(state, TT_KW, KW_END) == NULL) {
                // clang-format off
                if (
                    test_token_type(state, TT_PUSH_STR) || test_token_type(state, TT_PUSH_CSTR) || 
                    test_token_type(state, TT_PUSH_CHAR) || test_token_type(state, TT_PUSH_INT) || 
                    test_token_type(state, TT_PUSH_FLOAT) ||
                    test_token_type(state, TT_OP) || test_token_type(state, TT_IDENT)
                ) { // clang-format on
                    dynarray_pop(state->tokens, &state->curr_tok);
                    dynarray_push(tokens, state->curr_tok);
                }
            }
            if (!expect_token_type_ex(state, TT_KW, KW_END)) {
                return 1;
            }
            token_t* tokens_out = NULL;
            if (compile_value(state, tokens, &tokens_out) != 0) {
                return 1;
            }
            dynarray_destroy(tokens);
            if (dynarray_length(tokens_out) != 1) {
                log_error(&name.loc, "Constants can only have one value after calculation");
                return 1;
            }
            token_t value;
            dynarray_pop(tokens_out, &value);
            const_t c = {.name = name.str_v, .val = value};
            dynarray_push(state->prog.const_vars, c);
        }; break;
        case (KW_MEMORY): {
            token_t* tokens = dynarray_create(token_t);
            token_t name = *expect_token_type(state, TT_IDENT);
            while (test_token_type_ex(state, TT_KW, KW_END) == NULL) {
                // clang-format off
                if (test_token_type(state, TT_PUSH_INT) || test_token_type(state, TT_OP) || test_token_type(state, TT_IDENT) ) { // clang-format on
                    dynarray_pop(state->tokens, &state->curr_tok);
                    dynarray_push(tokens, state->curr_tok);
                }
            }
            if (!expect_token_type_ex(state, TT_KW, KW_END)) {
                return 1;
            }
            token_t* tokens_out = NULL;
            if (compile_value(state, tokens, &tokens_out) != 0) {
                return 1;
            }
            dynarray_destroy(tokens);
            if (dynarray_length(tokens_out) != 1 || tokens_out[0].type != TT_PUSH_INT) {
                log_error(&name.loc, "Memories can only have one Int value after calculation");
                return 1;
            }
            token_t* value;
            dynarray_pop(tokens_out, value);
            memory_t c = {.name = name.str_v, .size = value->int_v};
            dynarray_push(state->prog.memories, c);
        }; break;
        case (KW_FN): {
            token_t name = *expect_token_type(state, TT_IDENT);
            const char** in_args = dynarray_create(const char*);
            const char** out_args = dynarray_create(const char*);
            token_t tmp_tok = {0};
            expect_token_type_ex(state, TT_KW, KW_WITH);
            while (test_token_type_ex(state, TT_KW, KW_RETURNS) == NULL) {
                dynarray_pop(state->tokens, &tmp_tok);
                dynarray_push(in_args, tmp_tok.str_v);
            }
            expect_token_type_ex(state, TT_KW, KW_RETURNS);
            while (test_token_type_ex(state, TT_KW, KW_DO) == NULL) {
                dynarray_pop(state->tokens, &tmp_tok);
                dynarray_push(in_args, tmp_tok.str_v);
            }
            expect_token_type_ex(state, TT_KW, KW_DO);

            ast_op_t* body = parse_fnc_body(state);
            function_t fnc = {.loc = name.loc, .args = in_args, .return_args = out_args, .body=body, .name=name.str_v};
            dynarray_push(state->prog.funcs, fnc);
        }; break;
        default:
            assert(true && "TODO: parse all kw's");
        }
    }
    return 0;
}

int parse_item(parser_state_t* state, ast_op_t** body) {
    dynarray_pop(state->tokens, &state->curr_tok);
    switch (state->curr_tok.type) {
    case (TT_IDENT): {
        size_t id = 0;
        if ((id = find_const_id(state, state->curr_tok.str_v)) != -1) {
            ast_op_t v = {.loc = state->curr_tok.loc, .type = AOT_USE_CONST, .id = id};
            dynarray_push(*body, v);
        } else if ((id = find_mem_id(state, state->curr_tok.str_v)) != -1) {
            ast_op_t v = {.loc = state->curr_tok.loc, .type = AOT_USE_MEMORY, .id = id};
            dynarray_push(*body, v);
        } else if (find_func(state, state->curr_tok.str_v)) {
            ast_op_t v = {.loc = state->curr_tok.loc, .type = AOT_CALL_FUNC, .func_name = (char*)state->curr_tok.str_v};
            dynarray_push(*body, v);
        } else {
            log_error(&state->curr_tok.loc, "Unknown label '%s'", state->curr_tok.str_v);
            return 1;
        }
    }; break;
    case (TT_PUSH_BOOL):
    case (TT_PUSH_INT):
    case (TT_PUSH_FLOAT):
    case (TT_PUSH_CHAR):
    case (TT_PUSH_STR):
    case (TT_PUSH_CSTR):
    case (TT_OP): {
        ast_op_t v = {.loc = state->curr_tok.loc, .type = AOT_OP, .op = state->curr_tok};
        dynarray_push(*body, v);
    }; break;
    case (TT_KW): {
        switch (state->curr_tok.kw_type) {
        case (KW_WHILE): {
            ast_op_t* condition = dynarray_create(ast_op_t);
            ast_op_t* whilebody = dynarray_create(ast_op_t);
            while (test_token_type_ex(state, TT_KW, KW_DO) == NULL) {
                dynarray_pop(state->tokens, &state->curr_tok);
                if (parse_item(state, &condition)) {
                    return 1;
                }
            }
            assert(expect_token_type_ex(state, TT_KW, KW_DO));
            while (test_token_type_ex(state, TT_KW, KW_END) == NULL) {
                dynarray_pop(state->tokens, &state->curr_tok);

                if (parse_item(state, &whilebody)) {
                    return 1;
                }
            }
            assert(expect_token_type_ex(state, TT_KW, KW_END));
            ast_while_stat_t ws = {.body = whilebody, .condition = condition};
            ast_op_t v = {.loc = state->curr_tok.loc, .type = AOT_WHILE, .while_stat = ws};
            dynarray_push(body, v);
        }; break;
        case (KW_IF): {
            ast_op_t* condition = dynarray_create(ast_op_t);
            ast_op_t* ifbody = dynarray_create(ast_op_t);
            ast_op_t* else_body = dynarray_create(ast_op_t);
            bool is_else = false;
            bool is_elseif = false;
            while (test_token_type_ex(state, TT_KW, KW_DO) == NULL) {
                dynarray_pop(state->tokens, &state->curr_tok);
                if (parse_item(state, &condition)) {
                    return 1;
                }
            }
            assert(expect_token_type_ex(state, TT_KW, KW_DO));
            while (test_token_type_ex(state, TT_KW, KW_END) == NULL) {
                dynarray_pop(state->tokens, &state->curr_tok);

                if (parse_item(state, &ifbody)) {
                    return 1;
                }
                if (test_token_type_ex(state, TT_KW, KW_ELSE)) {
                    assert(expect_token_type_ex(state, TT_KW, KW_ELSE));
                    is_else = true;
                    if (test_token_type_ex(state, TT_KW, KW_IF)) {
                        is_elseif = true;
                        is_else = false;
                        break;
                    }
                }
            }
            if (is_else) {
                while (test_token_type_ex(state, TT_KW, KW_END) == NULL) {
                    dynarray_pop(state->tokens, &state->curr_tok);
                    if (parse_item(state, &else_body)) {
                        return 1;
                    }
                }
                assert(expect_token_type_ex(state, TT_KW, KW_END));
                ast_if_stat_t is = {.body = ifbody, .is_elseif = false, .condition = condition, .else_body = else_body};
                ast_op_t v = {.loc = state->curr_tok.loc, .type = AOT_IF, .if_stat = is};
                dynarray_push(body, v);
            } else if (is_elseif) {
                ast_op_t* tmp_body = dynarray_create(ast_op_t);
                parse_item(state, &tmp_body);

                ast_if_stat_t* elif_branch = malloc(sizeof(ast_if_stat_t));
                memcpy(elif_branch, &tmp_body[0].if_stat, sizeof(ast_if_stat_t));

                ast_if_stat_t is = {.body = ifbody, .is_elseif = true, .condition = condition, .elseif = elif_branch};
                ast_op_t v = {.loc = state->curr_tok.loc, .type = AOT_IF, .if_stat = is};
                dynarray_push(body, v);
            } else {
                assert(expect_token_type_ex(state, TT_KW, KW_END));
                ast_if_stat_t is = {.body = ifbody, .condition = condition};
                ast_op_t v = {.loc = state->curr_tok.loc, .type = AOT_IF, .if_stat = is};
                dynarray_push(body, v);
            }
        }; break;

        default: {
            log_error(&state->curr_tok.loc, "The %s is not supported inside function bodies", get_tok_str_dbg(&state->curr_tok));
            return 1;
        }; break;
        }
    }; break;
    }
    return 0;
}

ast_op_t* parse_fnc_body(parser_state_t* state) {
    ast_op_t* body = _dynarray_create(2, sizeof(ast_op_t));

    while (dynarray_length(state->tokens)) {
        if (test_token_type_ex(state, TT_KW, KW_END)) break;
        if (parse_item(state, &body)) {
            return NULL;
        }
    }
    return body;
}
